#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netinet/if_ether.h>
#ifdef LINUX
#include <netinet/ether.h>
#endif
#include <netinet/ip.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include <pthread.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pcap/pcap.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <ifaddrs.h>

#include "core.h"
#include "procMonitor.h"
#include "distributedMonitor.h"
#include "reportProcessor.h"
#include "uthash.h" //Hashtable code from http://uthash.sourceforge.net

#define MAX_BUF_SIZE	50

//-----------------------------
long int byteRx;
int nrRx;
long int byteTx;
int nrTx;
long int packRx;
long int packTx;

//-----------------

//GLOBAL Variables
int ports[65535]; //global socket table:  x -> nr of sockets associated with that port
int portsClosed[30]; //global table with sockets that were closed recently
int toexit; //global variable, that will be shared by all the files that will be equal to 1 when is to stop monitoring

FILE *fileoutput = NULL; //file where packets will be writed
char* filename; //path for fileoutput
int fileoutputsize = 0; //size that was already writen in fileoutput
pcap_t* descr = NULL; //pcap descriptor
char* filter = NULL; //filter to be used in pcap capture

pthread_mutex_t mutexTable = PTHREAD_MUTEX_INITIALIZER; //mutex for the read/write at global socket table
pthread_mutex_t mutexPacotes = PTHREAD_MUTEX_INITIALIZER; //mutex for the insertion/processing of list of packets
pthread_mutex_t mutexReport = PTHREAD_MUTEX_INITIALIZER; //mutex for read/write on report file where packets are stored

int reportCounter = 1;
int counterPacotes;
int ipcpipe;

nodeData* headNode; //head of the list where the packets captured will be stored before checking if is to write to fileoutput, or discard
nodeData* tailNode; //tail of the list

//declaration of the various threads that will be running while monitoring is active
pthread_t updateTable, filterProcess, capture, monitorProc, remoteReport, serverCommModule;
int initSuccess = 0;

//argumentos de utilização
int pid = -1;
int port = -1;
int server; //indicates that is the server or not, to lauch and stop the thread
int saveAll = 0; //indicates that if the packet is not from process will be saved in other file
FILE *saveAllFile = NULL; //file where packets not belong the process will be writed

struct in_addr* ipHost; //host ip
struct in_addr* ownIp; //own ip

struct hashIpStr *remoteAddress = NULL; /* important! initialize to NULL */

//--------Variables defined in distributedMonitor.c
extern char* device;
extern char* pipeFile;
extern char* socketFile;
extern char* hostPort;
extern char* saveLocation;
extern int remoteMonitorTime;
extern int remoteMonitor;
extern int pcapTime;
extern int pcapSize;

int final;

extern uid_t userID;
extern int userIDChange;

//--------Variable defined in reportProcessor.c
extern struct timeval hostCapStart; // time that the capture started. defined in the next module.

//----Hashtable Functions----

struct hashIpStr {
    int id; // key
    in_addr_t value;
    UT_hash_handle hh; // makes this structure hashable
};

//----------Start of Functions CODE---------------------------------

/*
 * Add one IP to the Hastable
 */
int add_ip(in_addr_t ip) {
    struct hashIpStr *s;

    s = malloc(sizeof (struct hashIpStr));
    s->id = (int) ip;
    s->value = ip;
    HASH_ADD_INT(remoteAddress, id, s); // id: name of key field
    return s->id;
}

/*
 * Find the IP in the Hastable
 */
struct hashIpStr *find_ip(int user_id) {
    struct hashIpStr *s;

    HASH_FIND_INT(remoteAddress, &user_id, s); //s: output pointer
    return s;
}
//---------------------------

/*
 * Insert one packet with respective header in the tail of the LinkedList nodeData
 */
int insertEnd(const struct pcap_pkthdr* pkthdr, const u_char* packet) {
    nodeData* node = malloc(sizeof (nodeData));

    node->pkthdr = malloc(sizeof (struct pcap_pkthdr));
    memcpy(node->pkthdr, pkthdr, sizeof (struct pcap_pkthdr));
    node->packet = malloc(pkthdr->caplen);
    memcpy(node->packet, packet, pkthdr->caplen);

    node->next = NULL;

    if (tailNode != NULL)
        tailNode->next = node;

    tailNode = node;

    if (headNode == NULL)
        headNode = node;

    counterPacotes++;

    return 0;
}

/*
 * Function will be used as CallBack function in pcap_loop operation, and this will be called to insert
 * one packet at time into the LinkedList using the function above.
 * It trys to lock the mutexPacotes so no one is using that list. At end unlock the mutexPacotes
 */
void packetprocessing(u_char *useless, const struct pcap_pkthdr* pkthdr, const u_char* packet) {
    pthread_mutex_lock(&mutexPacotes);
    insertEnd(pkthdr, packet);
    pthread_mutex_unlock(&mutexPacotes);
}

/*
 * function that will be run in thread, that will read text from the pipe file
 * and update the global socket table accordly.
 * If read '1 XXXXX' -> will increase the number in position XXXXX by 1
 * If read '0 XXXXX' -> will decrese the number in position XXXXX by 1
 * If read '2 0' -> will close the file descriptor that is reading the pipe, and at
 * 					that moment the variable toexit will be equal to 1.
 */
void *updateTableFunction(void *arg) {
    int fd;
    int numread;
    char buf[MAX_BUF_SIZE];
    int cmd, port;

    //Open the pipe for reading
    fd = open(pipeFile, O_RDONLY);

    while (1) {
        if (toexit == 1)
            break;

        numread = 0;

        // Read from the pipe
        memset(buf, 0, sizeof (buf));
        numread = read(fd, buf, sizeof (buf));
        if (numread <= 0) {
            close(fd);
            fd = open(pipeFile, O_RDONLY);

            numread = read(fd, buf, MAX_BUF_SIZE);
            if (numread == -1) {
                errorHandler("ERROR in update socket table - CANNOT READ FROM PIPE FILE");
                return 0;
            }
        }

        sscanf(buf, "%d %d", &cmd, &port);

        if (port <= 65535 && port >= 0) {
            pthread_mutex_lock(&mutexTable);
            if (cmd == 1) //add
            {
                if (ports[port] == -1)
                    ports[port] = 1;
                else
                    ports[port]++;
            } else
                if (cmd == 0) //remove
            {
                ports[port]--;

                //if that port is not used, there maybe packets captured that were captured when that port was still in use
                if (ports[port] == 0)
                    ports[port] = -1;
            } else
                if (cmd == 2) //exit command
                close(fd);

            pthread_mutex_unlock(&mutexTable);
        }
    }

    printf("EXIT THREAD UPDATE TABLE\n");
    fflush(stdout);

    close(fd);
    return 0;
}

/*
 * This function will be executed in thread and will prepare, process the filter and launch the packet capture
 * using the function packetprocessing, described above, to process each packet. It opens the first ouput file.
 * The interval that packets will be copied to user-space and process each one is defined by variable PCAPTIME.
 * The size of each packet to be captured will be defined by variable PCAPSIZE.
 * Both variables are defined in Settings file, readed by distributedMonitor.c
 */
void *captureFunction(void *arg) {
    char errbuf[PCAP_ERRBUF_SIZE];
    char* filter = NULL;
    struct bpf_program fp;
    bpf_u_int32 maskp;
    bpf_u_int32 netp;
    printf("before lookup capture\n");
    fflush(stdout);
    pcap_lookupnet(device, &netp, &maskp, errbuf);
    printf("after lookup\n");
    fflush(stdout);
    //open the device to capture
    descr = pcap_open_live(device, pcapSize, 0, pcapTime, errbuf);

    //change the execution permitions for real user
    setreuid(userID, userID);
    userIDChange = 1;

    if (descr == NULL) {
        char err[200];
        sprintf(err, "Couldn't open device %s: %s\n", device, errbuf);
        errorHandler(err);
        return 0;
    }

    if (arg == NULL) {
        printf("THERE IS NO FILTER\n");
        fflush(stdout);
    } else {
        filter = malloc(sizeof (char) * strlen((char*) arg));
        memset(filter, 0, strlen((char*) arg));
        strncpy(filter, (char*) arg + 1, strlen((char*) arg) - 2);
        printf("FILTER IN USE: %s\n", filter);
        fflush(stdout);

        //compilation of the filter
        if (pcap_compile(descr, &fp, filter, 0, netp) == -1) {
            char err[200];
            sprintf(err, "Couldn't parse filter %s: %s\n", filter, pcap_geterr(descr));
            errorHandler(err);
            return 0;
        }

        //attribuition of the filter to descriptor
        if (pcap_setfilter(descr, &fp) == -1) {
            char err[200];
            sprintf(err, "Couldn't install filter %s: %s\n", filter, pcap_geterr(descr));
            errorHandler(err);
            return 0;
        }
    }

    printf("capture from interface ->%s\n", device);
    fflush(stdout);

    fileoutput = (FILE*) pcap_dump_open(descr, "/tmp/remoteMonitor/filter/filtered0.dat");

    if (saveAll > 0) {
        saveAllFile = (FILE*) pcap_dump_open(descr, "/tmp/remoteMonitor/saveAll.dat");
    }

    char ftemp[100];
    sprintf(ftemp, "/tmp/remoteMonitor/filter/filtered0.dat");
    filename = ftemp;

    gettimeofday(&hostCapStart, NULL); //Get the time when capture will start

    pcap_loop(descr, -1, packetprocessing, NULL);

    printf("EXIT THREAD CAPTURA\n");
    fflush(stdout);

    return 0;
}

void* sshCommandThread(void* args) {
    char cmd[500];
    /* Guarantees that thread resources are deallocated upon return */
    pthread_detach(pthread_self());

    int sizen = sizeof (char) *500;
    memcpy(&cmd, args, sizen);


    system(cmd);
    return NULL;
}

/*
 * Function used only when the monitoring is set to be remote, to launch the monitoring in some remote machine
 * that we only know the IP address and the port from where the packet come.
 * Before executes the ssh command to start the remote monitoring, it checks the Hashtable to see if it has
 * that IP address, if the find_ip returns sucess, it don't execute the ssh command, because it was already launched
 * before. Only one try per IP address.
 */
void processPacketAddress(struct in_addr* socketAddr, int socketPort) {
    if (find_ip((int) socketAddr->s_addr) != NULL)
        return;

    add_ip(socketAddr->s_addr);

    char comand[500];

    char ffinal[20];

    //need to store the ipHost in string, since inet_ntoa() uses allways the same buffer
    char hostIP[30];
    strcpy(hostIP, inet_ntoa(*ipHost));

    if (final != 0)
        sprintf(ffinal, "-final");
    else
        if (final == 0)
        sprintf(ffinal, " ");

    if (filter != NULL)
        sprintf(comand, "ssh -f %s '/home/arknot/Desktop/captura/src/monitor -remote %s -interface %s -port %d -ip %s -saveAll -filter %s </dev/null >/tmp/output 2>&1 &'", inet_ntoa(*socketAddr), ffinal, inet_ntoa(*socketAddr), socketPort, hostIP, filter);
    else sprintf(comand, "ssh -f %s '/home/arknot/Desktop/captura/src/monitor -remote %s -interface %s -port %d -ip %s -saveAll </dev/null >/tmp/output 2>&1 &'", inet_ntoa(*socketAddr), ffinal, inet_ntoa(*socketAddr), socketPort, hostIP);

    printf("a invocar comando: %s\n", comand);
    fflush(stdout);

    //system(comand);
    pthread_t threadID;

    /* Create client thread */
    if (pthread_create(&threadID, NULL, sshCommandThread, (void *) &comand) != 0) {
        printf("pthread_create() failed");
        fflush(stdout);
    }

}

/*
 * Function that will run in thread and will process the LinkedList with the packets that were captured.
 * By using double buffering, it locks mutexPacotes mutex, and copy the header of the list, and puts
 * the global header and tail to NULL so when another packet is captured, new list will be started.
 * Will process the list, by looking at global socket table, and checking if that packet is to save
 * in file or to discard.
 * Before start the processing, it locks the mutexTable.
 * The intervale between processing the packets is 2 seconds.
 */
void *filterProcessFunction(void *arg) {
    nodeData* curr;
    nodeData* toDestroy;
    struct pcap_pkthdr* pkthdr;
    struct ip *ipc;
    struct tcphdr* tcphead;
    struct udphdr* udphead;
    u_char* packet;
    int portsClosedSize;

    int portSrc, portDst;

    while (1) {
        if (toexit == 1)
            break;

        msleep(2000);

        portsClosedSize = 0;

        //replace header and tail of the current list by NULL so new one can start
        pthread_mutex_lock(&mutexPacotes);
        curr = headNode;
        headNode = NULL;
        tailNode = NULL;
        counterPacotes = 0;
        pthread_mutex_unlock(&mutexPacotes);

        //start processing each packet in previous list
        //pthread_mutex_lock(&mutexTable);
        while (curr != NULL) {
            pkthdr = curr->pkthdr;
            packet = curr->packet;
            ipc = (struct ip*) (packet + sizeof (struct ether_header));

            switch (ipc->ip_p) {
                case 6:
                {
                    //TCP
                    tcphead = (struct tcphdr*) (packet + sizeof (struct ether_header) + sizeof (struct ip));
#ifdef LINUX
                    portSrc = ntohs(tcphead->source);
                    portDst = ntohs(tcphead->dest);
#else
                    portSrc = ntohs(tcphead->th_sport);
                    portDst = ntohs(tcphead->th_dport);    
#endif
                    break;
                }

                case 17:
                {
                    //UDP
                    udphead = (struct udphdr*) (packet + sizeof (struct ether_header) + sizeof (struct ip));
#ifdef LINUX
                    portSrc = ntohs(udphead->source);
                    portDst = ntohs(udphead->dest);
#else
                    portSrc = ntohs(udphead->uh_sport);
                    portDst = ntohs(udphead->uh_dport);   
#endif
                    break;
                }
                default:
                {
                    portSrc == -1;
                    portDst == -1;
                }
            }

            //check the range of the ports contain in the packets
            if (portSrc <= 65535 && portSrc >= 0 && portDst <= 65535 && portDst >= 0 && (ipc->ip_p == 17 || ipc->ip_p == 6)) {
                if ((ipc->ip_src.s_addr == ownIp->s_addr && ports[portSrc] != 0) || (ipc->ip_dst.s_addr == ownIp->s_addr && ports[portDst] != 0)) {
                    //check if remoteMonitor is active or not
                    if (remoteMonitor != 0) {
                        if (ipc->ip_src.s_addr == ownIp->s_addr)
                            processPacketAddress(&ipc->ip_dst, portDst);
                        else
                            if (ipc->ip_dst.s_addr == ownIp->s_addr)
                            processPacketAddress(&ipc->ip_src, portSrc);
                    }

                    if (ipc->ip_src.s_addr == ownIp->s_addr) {
                        byteTx += pkthdr->len;
                        packTx++;

                        if (byteTx > 2000000000) {
                            byteTx = 0;
                            nrTx++;
                        }
                    } else {
                        byteRx += pkthdr->len;
                        packRx++;

                        if (byteRx > 2000000000) {
                            byteRx = 0;
                            nrRx++;
                        }
                    }


                    //writes the packet and his pcap header in the file output
                    struct pcap_pkthdr sf_hdr;
                    sf_hdr.ts.tv_sec = pkthdr->ts.tv_sec;
                    sf_hdr.ts.tv_usec = pkthdr->ts.tv_usec;
                    sf_hdr.caplen = pkthdr->caplen;
                    sf_hdr.len = pkthdr->len;

                    pthread_mutex_lock(&mutexReport);
                    (void) fwrite(&sf_hdr, sizeof (sf_hdr), 1, fileoutput);
                    (void) fwrite((char *) packet, pkthdr->caplen, 1, fileoutput);
                    fileoutputsize += sizeof (sf_hdr) + pkthdr->caplen;
                    pthread_mutex_unlock(&mutexReport);

                    if (ports[portSrc] == -1 && portsClosedSize < 30) {
                        portsClosed[portsClosedSize] = portSrc;
                        portsClosedSize++;
                    } else
                        if (ports[portDst] == -1 && portsClosedSize < 30) {
                        portsClosed[portsClosedSize] = portDst;
                        portsClosedSize++;
                    }
                }//will be saved if not belong to process
                else
                    if (saveAll != 0) {
                    //writes the packet and his pcap header in the file output
                    struct pcap_pkthdr sf_hdr;
                    sf_hdr.ts.tv_sec = pkthdr->ts.tv_sec;
                    sf_hdr.ts.tv_usec = pkthdr->ts.tv_usec;
                    sf_hdr.caplen = pkthdr->caplen;
                    sf_hdr.len = pkthdr->len;

                    (void) fwrite(&sf_hdr, sizeof (sf_hdr), 1, saveAllFile);
                    (void) fwrite((char *) packet, pkthdr->caplen, 1, saveAllFile);
                }
            }

            toDestroy = curr;
            curr = toDestroy->next;

            free(toDestroy->pkthdr);
            free(toDestroy->packet);
            free(toDestroy);
        }

        //clear the temporary capture for sockets recently closed

        if (portsClosedSize > 0) {
            int i;
            int tt;
            pthread_mutex_lock(&mutexTable);
            for (i = 0; i < portsClosedSize; i++) {
                tt = portsClosed[i];
                ports[tt] = 0;
                portsClosed[i] = 0;
            }
            pthread_mutex_unlock(&mutexTable);
            portsClosedSize = 0;
        }

        fflush(fileoutput);

        //	pthread_mutex_unlock(&mutexTable);
    }

    printf("PackRx>%ld\n", packRx);
    printf("byteRx>%ld x %d\n", byteRx, nrRx);
    printf("PackTx>%ld\n", packTx);
    printf("byteTx>%ld x %d\n", byteTx, nrTx);
    printf("EXIT THREAD FILTERPROCESS\n");
    fflush(stdout);
    return 0;
}

/*
 * Sends the file that has the last captured packets to the Host, and
 * start one new file, to store the next packets until the next send.
 * Each file will be stored in /tmp/remoteMonitor/filter directory and will
 * be named filteredXXX.dat where the XXX will be the number of the report,
 * that will be incremented in each send
 */
void* sendRemoteToHost(void* bla) {
    char fname[100];
    char oldfname[100];
    int newfile;
    while (1) {
        msleep(remoteMonitorTime);
        newfile = 0;
        //case final is not defined, or defined but didn't registed with server at host
        if (final <= 0) {
            //case final is not defined
            if (final == 0) {
                pthread_mutex_lock(&mutexReport);
                if (fileoutputsize > 0) {
                    strcpy(oldfname, filename);
                    fclose(fileoutput); //close because the FILE point is created by pcap, and don't have permition to read from there
                    sprintf(fname, "/tmp/remoteMonitor/filter/filtered%d.dat", reportCounter);
                    fileoutput = (FILE*) pcap_dump_open(descr, fname);
                    filename = fname;
                    fileoutputsize = 0;
                    reportCounter++;
                    newfile = 1;
                }
                pthread_mutex_unlock(&mutexReport);
            }

            if (newfile == 1 || final == -1) {
                printf("sending\n");
                fflush(stdout);
                sendReport(ipHost, oldfname);
            }
        }

        if (toexit == 1)
            break;
    }
    printf("check if is final %d\n", final);
    fflush(stdout);

    //case its final send and already registed with server
    if (final == 1) {
        sendReport(ipHost, filename);
    }

    printf("EXIT THREAD REMOTEREPORT\n");
    fflush(stdout);
    //	sendFinalPacket(ipHost);
    return 0;
}

/*
 * Checks if the file where the preload is writing exits,
 * If is there, it will read the file, update the global socket table according with
 * the information readed and remove the file. Preload library will after that start
 * to write in the Pipe file.
 */
int checkPreloadFile() {
    int numread;
    char buf[MAX_BUF_SIZE];
    int cmd, port;

    int preloadFile = open(socketFile, O_RDONLY);

    if (preloadFile > 0) //file found
    {
        while (1) {
            numread = read(preloadFile, buf, sizeof (buf));
            if (numread == 0)
                break;
            if (numread > 0) {
                sscanf(buf, "%d %d\n", &cmd, &port);

                if (port <= 65535 && port >= 0) {
                    if (cmd == 1) //add
                        ports[port]++;
                    else
                        if (cmd == 0) //remove
                        ports[port]--;
                }
            }
        }
        close(preloadFile);
        remove(socketFile);

        return 1;
    }

    return 0;
}

/*
 * This function will receive the arguments passed by the distributedMonitor.c and process it
 * lauching then the
 */
void initCaptura(int argc, char *argv[]) {
    packRx = 0;
    packTx = 0;
    byteRx = 0;
    byteTx = 0;
    final = 0;

    struct ifaddrs * ifAddrStruct = NULL;
    void *arg = NULL;
    int args = 1;
    char* interfaceip = NULL;

    server = 0;

    ownIp = malloc(sizeof (struct in_addr));
    ipHost = malloc(sizeof (struct in_addr));

    if (argc < 2 && argv != NULL) {
        errorHandler("ERROR IN ARGUMENTS\n");
        return;
    }

    //checks the arguments passed
    while (args < argc && argv != NULL) {
        if (strcmp(argv[args], "-server") == 0) {
            server = 1; //sets the server variable so the process know where to get the HOST IP
            args++;
        } else
            if (strcmp(argv[args], "-pid") == 0 && args + 1 < argc)
            pid = atoi(argv[++args]);
        else
            if (strcmp(argv[args], "-port") == 0 && args + 1 < argc)
            port = atoi(argv[++args]);
        else
            if (strcmp(argv[args], "-ip") == 0 && args + 1 < argc)
            ipHost->s_addr = inet_addr(argv[++args]);
        else
            if (strcmp(argv[args], "-filter") == 0 && args + 1 < argc)
            filter = argv[++args];
        else
            if (strcmp(argv[args], "-remote") == 0 && args < argc) {
            remoteMonitor = 1;
            args++;
        } else
            if (strcmp(argv[args], "-saveAll") == 0 && args < argc) {
            saveAll = 1;
            args++;
        } else
            if (strcmp(argv[args], "-interface") == 0 && args + 1 < argc) {
            //interfaceip = malloc(sizeof(char)*30);
            interfaceip = argv[++args];
        } else
            if (strcmp(argv[args], "-final") == 0 && args < argc) {
            final = -1;
            args++;
        }            //redefinição das settings atraves da linha de comandos
        else
            if (strcmp(argv[args], "-device") == 0 && args < argc)
            device = argv[++args];
        else
            if (strcmp(argv[args], "-pipeFile") == 0 && args < argc)
            pipeFile = argv[++args];
        else
            if (strcmp(argv[args], "-socketFile") == 0 && args < argc)
            socketFile = argv[++args];
        else
            if (strcmp(argv[args], "-hostPort") == 0 && args < argc)
            hostPort = argv[++args];
        else
            if (strcmp(argv[args], "-remoteTime") == 0 && args < argc)
            remoteMonitorTime = atoi(argv[++args]);
        else
            if (strcmp(argv[args], "-saveLocation") == 0 && args < argc)
            saveLocation = argv[++args];
        else
            if (strcmp(argv[args], "-pcapTime") == 0 && args < argc)
            pcapTime = atoi(argv[++args]);
        else
            if (strcmp(argv[args], "-pcapSize") == 0 && args < argc)
            pcapSize = atoi(argv[++args]);
        else
            args++; //argument given not valid, incrementing args to try to exit
    }

    getifaddrs(&ifAddrStruct); //gets all the devices that can communicate in network
    while (ifAddrStruct != NULL) {
        if (interfaceip != NULL) {
            //Search for the device(interface) that is using some IP given when process is started
            if (ifAddrStruct->ifa_addr->sa_family == AF_INET) {
                ownIp->s_addr = ((struct sockaddr_in *) ifAddrStruct->ifa_addr)->sin_addr.s_addr;
                if (inet_addr(interfaceip) == ownIp->s_addr) {
                    device = ifAddrStruct->ifa_name;
                    break;
                }
            }
        } else
            //search for the interface that have the name given in the device
            if (ifAddrStruct->ifa_addr->sa_family == AF_INET && strcmp(ifAddrStruct->ifa_name, device) == 0) {
            ownIp->s_addr = ((struct sockaddr_in *) ifAddrStruct->ifa_addr)->sin_addr.s_addr;
        }
        ifAddrStruct = ifAddrStruct->ifa_next;
    }

    if (server == 1) {
        ipHost->s_addr = ownIp->s_addr; //if is the server his own ip is the Hosts IP
        pthread_create(&serverCommModule, NULL, communicationModule, NULL);
    }

    counterPacotes = 0;

    headNode = NULL;
    tailNode = NULL;

    toexit = 0;

    //Creates the pipe used to communicate betwen  the core and the process monitor (by pid or using LD_Preload)
    umask(011);
    if (mkfifo(pipeFile, 0666) == -1) {
        umask(022);
        errorHandler("CAN'T CREATE PIPE");
        return;
    } else {
        umask(022);
        chown(pipeFile, userID, -1); //since the efective owner is root
        int pps = checkPreloadFile();
        if (pps == 0) //Check is preload file was created
        {
            if (pid != -1) {
                //Thread that will monitoring the process that has the pid argument
                pthread_create(&monitorProc, NULL, pidMonitorThread, (void*) pid); //gives warning in 64bits
            } else
                if (port != -1) {
                //Thread that will monitoring the process that is using the port argument
                pthread_create(&monitorProc, NULL, socketMonitorThread, (void*) port); //gives warning in 64bits
            }
        } else {
            //TODO alterar aqui
            printf("PRELOAD FILE FOUND\n");
            fflush(stdout);
            port = -1;
            pid = -1;
        }


        //Thread that will read from pipe file and update the Global Table of IPs under monitorization
        pthread_create(&updateTable, NULL, updateTableFunction, arg);

        //Thread that will lauch the packet capture based in LibPCap library
        pthread_create(&capture, NULL, captureFunction, (void*) filter);

        //thread that will process the packets that were captured
        pthread_create(&filterProcess, NULL, filterProcessFunction, arg);

        pthread_create(&remoteReport, NULL, sendRemoteToHost, (void*) pid);

        initSuccess = 1;
    }
}

/*
 * Used by distributedMonitor main function to be waiting until all threats stop.
 * If that happens, it closes the fileoutput
 */
void joinThreads() {
    if (pid != -1 || port != -1)
        pthread_join(monitorProc, NULL);

    pthread_join(capture, NULL);
    pthread_join(updateTable, NULL);
    pthread_join(filterProcess, NULL);
    pthread_join(remoteReport, NULL);

    if (server == 1)
        pthread_join(serverCommModule, NULL);

    if (fileoutput != NULL)
        fclose(fileoutput);

    if (saveAll > 0)
        fclose(saveAllFile);
}

int toExitCapt() {
    if (initSuccess == 1 && toexit != 1) {
        toexit = 1; //global
        if (descr != NULL)
            pcap_breakloop(descr);
        //thread tratamento da tabela
        int fd = open(pipeFile, O_WRONLY);
        char* str = "2 0";
        write(fd, str, strlen(str));
        close(fd);

        if (server == 1)
            forceExistServer(ipHost);
    }
    return 1;
}
