#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/if_ether.h>
#ifdef LINUX
#include <netinet/ether.h>
#endif
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <pthread.h>
#include <dirent.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pcap.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <net/if.h>
#include <ifaddrs.h>
#include <time.h>
#include <signal.h>

#include "distributedMonitor.h"
#include "core.h"
#include "procMonitor.h"
#include "reportProcessor.h"


#define SETTINGS "settings"
#define MONDIR "/tmp/remoteMonitor"     // Where remote and local monitor data is saved
#define MAX_BUF_SIZE	50
#define MAXPENDING 		10    	// Maximum outstanding connection requests
#define RCVBUFSIZE 		60000  	// Size of receive/send buffer
#define REMOTECHECKSIZE 50   	// Size of receive/send remotebuffer
#define REMOTECHECKPORT 40004;  //port where the remoteMonitor server will listen

#define TIMEOUT         5;

char* device; //device from where the capture must happen
char* pipeFile; //path for the pipe file
char* socketFile; //path where the socket file produced by mysockets.so is
char* hostPort; //Host report server will be listen for incoming of new connections
int remoteMonitorTime; //time between the sends to the server about the local capture
int remoteMonitor = 0; //by default we don't do remote monitoring
char* saveLocation;
int pcapTime;
int pcapSize;

int dynamicGraphDraw = 0; //by default the graph is draw when all the monitoring is finished
extern int toDrawInt;

struct timeval startTime; // time that the capture started. defined in the next module.

pthread_t remoteCheck; //thread that will create one socket where the host will send message

int sock = -1; //socket for sending the reports
int sockOpened[100];

struct timeval times[100];
int sockOpenedCounter = 0;
int reportNr;

extern int toexit;
extern struct in_addr* ipHost;
int servSock = -1;

uid_t userID;
int userIDChange;

extern int final;

//------------------------------------------------------
int sockfd[100];
int sockRecv[100];
FILE* fileOpened[100];
int socksize;
int sockRecvSize;
pthread_mutex_t sockfdmutex = PTHREAD_MUTEX_INITIALIZER;
fd_set sockSet;
fd_set sockWrite;
struct timeval selTimeout;
int maxDescriptor;

//to control how many threads are receiving in the end
int inThread;
int outThread;

/* Structure of arguments to pass to client thread */
struct ThreadArgs {
    int clntSock; /* Socket descriptor for client */
    int position;
};


//------------------------------------------------------

//------------------ packet processing-------------------
extern struct timeval hostCapStart; // time that the capture started. defined in the next module.

void errorHandler(char* msg) {
    printf("ERROR DETECTED: %s\n", msg);
    perror(msg);
    fflush(stdout);
    toExitCapt();
}

int SetSocketTimeout(int connectSocket, int milliseconds) {
    struct timeval tv;

    tv.tv_sec = milliseconds / 1000;
    tv.tv_usec = (milliseconds % 1000) * 1000;

    return setsockopt(connectSocket, SOL_SOCKET, SO_RCVTIMEO, (char *) &tv, sizeof tv);
}

void *tcpReceiver(void *threadArgs) {
    int clntSock; /* Socket descriptor for client connection */
    int position;

    /* Guarantees that thread resources are deallocated upon return */
    pthread_detach(pthread_self());

    /* Extract socket file descriptor from argument */
    clntSock = ((struct ThreadArgs *) threadArgs) -> clntSock;
    position = ((struct ThreadArgs *) threadArgs) -> position;

    //search for file associated with socket
    FILE* filetemp;
    int k;
    for (k = 0; k < 100; k++) {
        if (sockfd[k] == clntSock) {
            filetemp = fileOpened[k];
            break;
        }
    }

    //prepares buffer
    u_char* buffer = malloc(sizeof (u_char*) * RCVBUFSIZE); /* Buffer for echo string */

    if (buffer == NULL) {
        printf("ERROR MALLOC TCPRECEIVER\n");
        fflush(stdout);
    }

    int recvMsgSize; /* Size of received message */
    memset(buffer, 0, sizeof (buffer));

    int toProcessDynamic = 0;
    //start receiving
    if (toexit == 1) {
        SetSocketTimeout(clntSock, 25000);
    }

    while ((recvMsgSize = recv(clntSock, buffer, RCVBUFSIZE, 0)) > 0) {
        // in order to get the information captured in remote pc, we need to analise the report
        if (dynamicGraphDraw == 1) {
            //	printf("in dynadraw ->>>>>>>>>>>>>>>>");
            if (ftell(filetemp) < 24) //first report received
                newGraphTrace(buffer, recvMsgSize, 1);
            else
                newGraphTrace(buffer, recvMsgSize, 0);
            toProcessDynamic = 1;
            //	printf("out dynadraw\n");
        }

        // Write message back to log
        fwrite(buffer, recvMsgSize, 1, filetemp);
        fflush(filetemp);
    }

    //check if there is something to process the report just received
    if (toProcessDynamic == 1)
        processGraphTrace(clntSock);

    if (toexit != 1)
        sockRecv[position] = 0; //remove from receiving array
    else {
        fclose(filetemp);
        close(clntSock);
        outThread++;
    }
    return (NULL);
}

void launchSocketThread(int sfd, int pos) {
    pthread_t threadID; /* Thread ID from pthread_create() */

    /* Create separate memory for client argument */
    struct ThreadArgs *threadArgs; /* Pointer to argument structure for thread */

    if ((threadArgs = (struct ThreadArgs *) malloc(sizeof (struct ThreadArgs))) == NULL) {
        printf("malloc() failed");
        fflush(stdout);
    }

    threadArgs -> clntSock = sfd;
    threadArgs -> position = pos;

    /* Create client thread */
    if (pthread_create(&threadID, NULL, tcpReceiver, (void *) threadArgs) != 0) {
        printf("pthread_create() failed");
        fflush(stdout);
    }
}

void* receiveReport(void* blah) /* TCP client handling function */ {
    int i = 0;
    int j = 0;
    int l;
    int toIns;
    int tmp;
    //vad-unused:    pthread_t threadID;              /* Thread ID from pthread_create() */

    inThread = 0;
    outThread = 0;

    printf("REPORT SERVER READY TO RECEIVE\n");
    fflush(stdout);
    while (1) {
        /* Timeout specification */
        /* This must be reset every time select() is called */
        selTimeout.tv_sec = TIMEOUT; /* timeout (secs.) */
        selTimeout.tv_usec = 0; /* 0 microseconds */

        /* Initialize maxDescriptor for use by select() */
        maxDescriptor = -1;

        /* Zero socket descriptor vector and set for server sockets */
        /* This must be reset every time select() is called */
        FD_ZERO(&sockSet);

        /* Create list of ports and sockets to handle ports */
        for (i = 0; i < socksize; i++) {
            tmp = sockfd[i];

            toIns = 0;
            for (j = 0; j < 100; j++)
                if (tmp == sockRecv[j]) {
                    toIns = 1;
                    break;
                }

            if (toIns == 0) {
                if (toexit == 1 && final == 1) {
                    inThread++;
                    launchSocketThread(tmp, 0);
                } else {
                    FD_SET(sockfd[i], &sockSet);

                    /* Determine if new descriptor is the largest */
                    if (sockfd[i] > maxDescriptor)
                        maxDescriptor = sockfd[i];
                }
            }
        }

        //if the exist sign is marked and no fd was added
        if (toexit == 1 && maxDescriptor == -1) {
            printf("going out from received select\n");
            //return 0;
            break;
        }

        /* Suspend program until descriptor is ready or timeout */
        if (select(maxDescriptor + 1, &sockSet, NULL, NULL, &selTimeout) != 0) {
            for (i = 0; i < socksize; i++) {
                if (FD_ISSET(sockfd[i], &sockSet)) {
                    int poss = -1;
                    for (l = 0; l < 100; l++) {
                        if (sockRecv[l] == 0) {
                            sockRecv[l] = sockfd[i];
                            poss = l;
                            break;
                        }
                    }
                    //launch thread to receive the report and store it. After receive, put socket again in list and call stop
                    //argumentos: socket e posição que ocupa no sockRecv

                    launchSocketThread(sockfd[i], poss);
                }
            }
        }
    }

    while (inThread > 0) {
        msleep(2000);
        printf("inThread %d vs %d outThread\n", inThread, outThread);
    }

    printf("EXIT REPORT CHECKER\n");
    return 0;
}

void launchRemoteStop() {
    int sock;
    struct sockaddr_in servAddr;
    unsigned short servPort;
    char *msgString;
    unsigned int msgStringLen;
    int socktemp;
    int i;

    for (i = 0; i < socksize; i++) {
        socktemp = sockfd[i];

        struct sockaddr_in remoty;
        int sizeRemoty = sizeof (remoty);

        getpeername(socktemp, (struct sockaddr*) &remoty, &sizeRemoty);

        if (remoty.sin_addr.s_addr != ipHost->s_addr) {
            msgString = malloc(sizeof (char) * REMOTECHECKSIZE);
            strcpy(msgString, "close");

            servPort = REMOTECHECKPORT;

            if ((sock = socket(AF_INET, SOCK_STREAM, 0)) > 0) {
                SetSocketTimeout(sock, 2000);

                memset(&servAddr, 0, sizeof (servAddr));
                servAddr.sin_family = AF_INET;
                servAddr.sin_addr.s_addr = remoty.sin_addr.s_addr;
                servAddr.sin_port = htons(servPort);

                if (connect(sock, (struct sockaddr *) &servAddr, sizeof (servAddr)) == 0) {
                    msgStringLen = strlen(msgString);
                    send(sock, msgString, msgStringLen, 0);
                }
            }
        }
    }
}

void* communicationModule(void* bla) {
    int clntSock;
    struct sockaddr_in servAddr;
    struct sockaddr_in clntAddr;
    unsigned short servPort;
    unsigned int clntLen;
    char filePath[100];
    pthread_t reportServer;

    servPort = atoi(hostPort);

    if ((servSock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        errorHandler("socket creation in communicationModule fail\n");
        return 0;
    }

    memset(&servAddr, 0, sizeof (servAddr));
    servAddr.sin_family = AF_INET;
    servAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servAddr.sin_port = htons(servPort);

    if (bind(servSock, (struct sockaddr *) &servAddr, sizeof (servAddr)) < 0) {
        errorHandler("bind in communicationModule fail\n");
        return 0;
    }

    if (listen(servSock, MAXPENDING) < 0) {
        errorHandler("listen in communicationModule fail\n");
        return 0;
    }

    socksize = 0;

    //lauch thread to receive the reports
    pthread_create(&reportServer, NULL, receiveReport, NULL);

    printf("SERVER READY TO RECEIVE\n");
    for (;;) {
        clntLen = sizeof (clntAddr);

        if ((clntSock = accept(servSock, (struct sockaddr *) &clntAddr, &clntLen)) < 0) {
            errorHandler("accept in communicationModule fail\n");
            return 0;
        }

        if (toexit == 1) {
            close(clntSock);
            break;
        }

        char* ip = inet_ntoa(clntAddr.sin_addr);
        printf("Handling client %s\n", ip);
        fflush(stdout);

        //start receive the timers
        struct timeval tm1;
        struct timeval tm2;
        struct timeval remoteStartTime;
        struct timeval recTime;
        char msg1[10];
        recv(clntSock, &remoteStartTime, sizeof (struct timeval), 0);
        sprintf(msg1, "givetime"); // ask for time
        send(clntSock, msg1, sizeof (msg1), 0);
        gettimeofday(&tm1, NULL);
        recv(clntSock, &recTime, sizeof (struct timeval), 0);
        gettimeofday(&tm2, NULL);
        addNewElement(clntSock, &remoteStartTime, &tm1, &tm2, &recTime, clntAddr.sin_addr.s_addr);

        //-----------------------------------
        SetSocketTimeout(clntSock, 2000);
        sockfd[socksize] = clntSock; //add socket to array
        sprintf(filePath, MONDIR"/reports/report%d.dat", clntSock);
        fileOpened[socksize] = fopen(filePath, "w+");
        socksize++;
    }

    printf("LAUNCHING STOP\n");
    launchRemoteStop();

    close(servSock);
    pthread_join(reportServer, NULL);
    printf("EXIT SERVER THREAD\n");
    fflush(stdout);
    return 0;
}

//void forceExistServer(char* hostIP){

void forceExistServer(struct in_addr* hostIP) {
    if (toexit == 1 && servSock != -1) {
        struct sockaddr_in echoServAddr; /* Echo server address */
        int sock = socket(AF_INET, SOCK_STREAM, 0);

        memset(&echoServAddr, 0, sizeof (echoServAddr)); /* Zero out structure */
        echoServAddr.sin_family = AF_INET; /* Internet address family */
        echoServAddr.sin_addr.s_addr = hostIP->s_addr; /* Server IP address */
        echoServAddr.sin_port = htons(atoi(hostPort)); /* Server port */

        /* Establish the connection to the echo server */
        connect(sock, (struct sockaddr *) &echoServAddr, sizeof (echoServAddr));
        close(sock);
    }
}

void* remoteCheckStop(void* unused) {
    int servSock;
    int clntSock;
    struct sockaddr_in servAddr;
    struct sockaddr_in clntAddr;
    unsigned int clntLen;

    if ((servSock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        errorHandler("socket() in remoteCheckStop failed");

    int port = REMOTECHECKPORT;

    memset(&servAddr, 0, sizeof (servAddr)); /* Zero out structure */
    servAddr.sin_family = AF_INET; /* Internet address family */
    servAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* Any incoming interface */
    servAddr.sin_port = htons(port); /* Local port */

    if (bind(servSock, (struct sockaddr *) &servAddr, sizeof (servAddr)) < 0)
        errorHandler("bind() in remoteCheckStop failed");

    if (listen(servSock, MAXPENDING) < 0)
        errorHandler("listen() in remoteCheckStop failed");

    for (;;) /* Run forever */ {
        clntLen = sizeof (clntAddr);

        if ((clntSock = accept(servSock, (struct sockaddr *) &clntAddr, &clntLen)) < 0)
            errorHandler("accept() in remoteCheckStop failed");

        char echoBuffer[REMOTECHECKSIZE]; /* Buffer for string */
        int recvMsgSize; /* Size of received message */

        /* Receive message from client */
        if ((recvMsgSize = recv(clntSock, echoBuffer, REMOTECHECKSIZE, 0)) < 0)
            errorHandler("recv() in remoteCheckStop failed");

        if (strcmp(echoBuffer, "close") == 0) {

            break;
        }

        close(clntSock); /* Close client socket */
    }
    close(servSock);
    stopHandler(0);
    return 0;
}

/*
 * This sends the report stored at filepath to the hostIP
 * At first time it connects
 */
int sendReport(struct in_addr* hostIP, char* filepath) {
    int port = atoi(hostPort);
    struct sockaddr_in servAddr;
    static int firstComm = 0;
    int readed = 0;
    int sended = 0;

    u_char* buffer = malloc(sizeof (u_char*) * RCVBUFSIZE);

    /* Create a reliable, stream socket using TCP */
    int counter = 0;
    while (sock == -1 && counter < 10) {
        printf("Socket Opening\n");
        if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
            errorHandler("socket creation in SendSocket fail\n");

        SetSocketTimeout(sock, 5000);

        /* Construct the server address structure */
        memset(&servAddr, 0, sizeof (servAddr)); /* Zero out structure */
        servAddr.sin_family = AF_INET; /* Internet address family */
        servAddr.sin_addr.s_addr = hostIP->s_addr; /* Server IP address */
        servAddr.sin_port = htons(port); /* Server port */

        /* Establish the connection to the echo server */
        if (connect(sock, (struct sockaddr *) &servAddr, sizeof (servAddr)) < 0) {
            printf("connect in SendSocket fail\n");
            close(sock);
            sock = -1;
        } else
            //sends the local time, when the capture started.
            if (send(sock, &hostCapStart, sizeof (struct timeval), 0) == -1) {
            printf("Error in sending the first report, it is possible that host finish\n");
            close(sock);
            sock = -1;
        } else {
            //connection done because pass all if's
            char msg1[100];
            recv(sock, msg1, sizeof (msg1), 0);
            struct timeval tm;
            gettimeofday(&tm, NULL);
            send(sock, &tm, sizeof (struct timeval), 0); //sends the current epoch time
            if (servSock == -1)
                pthread_create(&remoteCheck, NULL, remoteCheckStop, NULL);

            if (final == -1) {
                final = 1;
                return 0;
            }
        }
        counter++;
    }

    if (counter >= 10)
        errorHandler("ERROR SEND SOCKET");
    //--------------------------- read file and send -----------------------
    FILE* fnew;
    if ((fnew = fopen(filepath, "r")) == NULL) {
        printf("FILEEE DO ERROO %s\n", filepath);
        fflush(stdout);

        errorHandler("CAN'T OPEN FILE TO SEND");
        return 0;
    }
    if (firstComm == 0) //its the first report from that pc, so we need to send the file header (24 bytes)
    {
        readed = fread(buffer, 1, RCVBUFSIZE, fnew);
        firstComm = 1;
    } else {
        fseek(fnew, 24, SEEK_SET);
        readed = fread(buffer, 1, RCVBUFSIZE, fnew);
    }

    while (readed > 0) {
        /* Send the string to the server */
        sended = send(sock, buffer, readed, 0);
        if (sended != readed) //tests if the number of data sended is the same that wanted to send
        {
            if (sended == -1) {
                errorHandler("Error in sending the report, it is possible that host finish\n");
                return -1;
            }
            sock = -1;
            readed = 0;
        } else
            readed = fread(buffer, 1, RCVBUFSIZE, fnew);
    }
    free(buffer);
    fclose(fnew);

    return 0;
}

/*
 * This will read the settings file and retrieve the various arguments from it
 */
void loadSettings() {
    char cmd[100];
    char var[100];

    device = malloc(sizeof (char) *20);
    pipeFile = malloc(sizeof (char) *20);
    socketFile = malloc(sizeof (char) *20);
    hostPort = malloc(sizeof (char) *20);
    saveLocation = malloc(sizeof (char) *300);

    FILE * settingsfile = fopen(SETTINGS, "r");
    char buf[MAX_BUF_SIZE];

    if (access(SETTINGS, F_OK) == 0 && settingsfile > 0) {
        while (fgets(buf, sizeof buf, settingsfile) != NULL) {
            sscanf(buf, "%s %s", cmd, var);

            if (strcmp(cmd, "device") == 0) {
                strcpy(device, var);
                printf("device ->%s\n", device);
                fflush(stdout);
            } else
                if (strcmp(cmd, "pipe_name") == 0) {
                strcpy(pipeFile, var);
                printf("pipename ->%s\n", pipeFile);
                fflush(stdout);
            } else
                if (strcmp(cmd, "socket_file") == 0) {
                strcpy(socketFile, var);
                printf("socketFile ->%s\n", socketFile);
                fflush(stdout);
            } else
                if (strcmp(cmd, "hostPort") == 0) {
                strcpy(hostPort, var);
                printf("hostPort ->%s\n", hostPort);
                fflush(stdout);
            } else
                if (strcmp(cmd, "remoteTime") == 0) {
                remoteMonitorTime = atoi(var);
                printf("remoteMonitorTime ->%d\n", remoteMonitorTime);
                fflush(stdout);
            } else
                if (strcmp(cmd, "saveLocation") == 0) {
                strcpy(saveLocation, var);
                printf("saveLocation ->%s\n", saveLocation);
                fflush(stdout);
            } else
                if (strcmp(cmd, "pcapTime") == 0) {
                pcapTime = atoi(var);
                printf("pcapTime ->%d\n", pcapTime);
                fflush(stdout);
            } else
                if (strcmp(cmd, "pcapSize") == 0) {
                pcapSize = atoi(var);
                printf("pcapSize ->%d\n", pcapSize);
                fflush(stdout);
            }
        }
    } else {
        printf("Ficheiro Settings não encontrado. Usando pre-definicoes.\n");
        fflush(stdout);
#ifdef LINUX
        sprintf(device, "eth0");
#else
        sprintf(device, "en0");
#endif
        sprintf(pipeFile, MONDIR"/pipe");
        sprintf(socketFile, "/tmp/socketPorts");
        sprintf(hostPort, "23332");
        remoteMonitorTime = 10000; //envios a cada 10segundos
        sprintf(saveLocation, "/tmp/monitorFinal");
        pcapTime = 1000; //captura a cada 1 segundos
        pcapSize = 100; //capturado 100bytes do pacote
    }
    mkdir("/tmp/monitorFinal", 0777);
    chown("/tmp/monitorFinal", userID, -1);
}

/*
 * To start monitoring from graphic interface, this is the function that need to be called
 * If pid > 0, then it will start monitoring based on pid, else will be waiting for someone write in pipe file
 */
void startMonitoring(int pid, int port, int remote, int dynamic, const char* filter) {
    char* argv[50];
    int argc = 1;
    int i;

    //vad
    fprintf(stderr, ">> startMonitoring(pid=%d,port=%d,remote=%d,dynamic=%d,filter=%s)\n",
            pid, port, remote, dynamic, filter);

    //build array with args to simulate the start by console
    argv[argc] = malloc(sizeof (char) *10);
    sprintf(argv[argc], "-server");
    argc++;

    if (pid > 0) {
        argv[argc] = malloc(sizeof (char) *5);
        sprintf(argv[argc], "-pid");
        argc++;
        argv[argc] = malloc(sizeof (char) *10);
        sprintf(argv[argc], "%d", pid);
        argc++;
    } else
        if (port > 0) {
        argv[argc] = malloc(sizeof (char) *5);
        sprintf(argv[argc], "-port");
        argc++;
        argv[argc] = malloc(sizeof (char) *10);
        sprintf(argv[argc], "%d", port);
        argc++;
    }

    if (strlen(filter) > 0) {
        printf("adding filter\n");
        argv[argc] = malloc(sizeof (char) *10);
        sprintf(argv[argc], "-filter");
        argc++;
        argv[argc] = malloc(sizeof (char) * strlen(filter));
        sprintf(argv[argc], "'%s'", filter);
        argc++;
    }

    remoteMonitor = remote;
    dynamicGraphDraw = dynamic;

    setInterfaceInteraction(); //sets the processing of graph information

    main(argc, argv);

    for (i = 1; i < argc; i++)
        free(argv[i]);

}

/*
 * Function to be called when to stop the monitoring, by error
 * Will call the function in core.c
 */
void stopMonitoring() {
    //call the exit function that will stop all the threads and put the toExit variable equal to 1
    toExitCapt();
}

/*
 * Function that exits from the process but calls stopMonitoring first
 */
void stopHandler(int unused) {
    stopMonitoring();
}

/***************************************************************************
 * Main function, used for launch the monitoring in the console
 */
int main(int argc, char *argv[]) {

    //check if there is another execution already on
    int ff = mkdir(MONDIR, 0777);
    if (ff != 0) {
        perror(MONDIR);
        printf("Monitoring already running?\n");
        return 1;
    }

    printf("getuid %d ---- geteuid %d\n", getuid(), geteuid());

    userID = getuid();
    userIDChange = userID != geteuid();

    //grab the SIGINT so when user want to stop monitoring by using the command Ctrl + C
    //it stops the monitoring and stop all the threads and write all the files
    signal(SIGINT, stopHandler);

    //creates the folders, for reports received by host, and the packets captured
    if (mkdir(MONDIR"/reports", 0777) == -1 || mkdir(MONDIR"/filter", 0777) == -1) {
        errorHandler("cannot create report/filter directories\n");
        return 1;
    }

    if ( userIDChange ) {
        //change user id from the directories created above
        chown(MONDIR, userID, -1);
        chown(MONDIR"/reports", userID, -1);
        chown(MONDIR"/filter", userID, -1);
    }

    reportNr = 0;
    loadSettings();

    //vad
    fprintf(stderr, ">> initCaptura(argc=%d, ...)\n", argc);

    initCaptura(argc, argv);
    //vad
    fprintf(stderr, ">> joinThreads() >>\n");

    joinThreads();

    //----------Exiting-----------
    //vad
    fprintf(stderr, ">> Exiting >>\n");

    //if is the server, process all the traces received and create the single ordered report
    if (servSock != -1) {
        printf("GOING TO START CALCULATE TIMES\n");
        fflush(stdout);
        calculateTimes();
        insertOrderedPackets();

        //saves the temporary files to saveLocation
        DIR *dp;
        struct dirent *ep;

        dp = opendir(saveLocation);
        int max = 0;
        int temp = 0;

        //find the max value at folder
        if (dp != NULL) {
            while ((ep = readdir(dp))) {
                temp = atoi(ep->d_name);
                if (temp > max)
                    max = temp;
            }
            (void) closedir(dp);

            char fullpath[300];
            sprintf(fullpath, "%s/%d", saveLocation, max + 1);
            if (mkdir(fullpath, 0777) == -1) {
                errorHandler("CANNOT CREATE NEW FOLDER AT SAVELOCATION PATH");
            } else {
                char command[300];
                sprintf(command, "cp -r "MONDIR"/* %s/", fullpath);
                system(command);
                printf("SAVED IN DIRECTORY: %s\n", fullpath);
                fflush(stdout);
            }
        } else errorHandler("cannot open the final directory\n");
    }

    if (toDrawInt != 1) {
        //removes the directory MONDIR, so other monitor can be launched
        system( "rm -r "MONDIR );
    }

    return 0;
}
