
#ifdef LINUX
  #ifndef RTLD_NEXT
  #define _GNU_SOURCE
  #endif
#endif

#include <dlfcn.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <unistd.h>

#define PIPE_LOCATION		"/tmp/remoteMonitor/pipe"
#define FILE_LOCATION		"/tmp/socketPorts"

#define MAX_BUF_SIZE	50

int fdOpen[256]; //limitado a   255 descritores abertos ao mesmo tempo:
            // -> 0 não atribuido ; x -> atribuido (porto); -1 -> desconhecido

int filefd = 0; //canal de escrita pipe ou file
int typefd = -1; //tipo de escrita actual, 0->pipe ; 1->file

void reportLocation() { 
    // vad:
    if (filefd > 0) printf("ESTRANHO: abrindo novamente o pipe/ficheiro!!\n");
    else fprintf(stderr,"DEBUG:  it works!!!\n");

    filefd = open(PIPE_LOCATION, O_WRONLY);
    if (filefd < 0) {
        perror("?? "PIPE_LOCATION);
        //PIPE não encontrado, vamos usar um ficheiro para log.
        remove(FILE_LOCATION);
        filefd = open(FILE_LOCATION, O_CREAT | O_RDWR | O_TRUNC, 0644);
        //chown(FILE_LOCATION, getuid(), -1);
        typefd = 1;
    } else typefd = 0;
}

void writeSocketPort(int add, int port) {
    char toadd[MAX_BUF_SIZE];

    // if first call, open pipe or file:
    if ( filefd == 0 || typefd == -1 )
        reportLocation();

    memset(toadd, 0, sizeof (toadd));
    if (typefd == 0) //writing in pipe
    {
        if (add == 0) sprintf(toadd, "%d %d", 0, port); // socket closed
        else sprintf(toadd, "%d %d", 1, port);  // socket opened
    } else
        if (typefd == 1) //writing in file
    {
        if (add == 0) sprintf(toadd, "%d %d\n", 0, port);
        else sprintf(toadd, "%d %d\n", 1, port);
    }

    //write(filefd, toadd, sizeof(toadd));
    int (*libc_write)(int, const void *, size_t);
    *(void **) (&libc_write) = dlsym(RTLD_NEXT, "write");
    (*libc_write)(filefd, toadd, sizeof (toadd));
}

int socket(int domain, int type, int protocol);

int socket(int domain, int type, int protocol) {
    int (*libc_socket)(int, int, int);
    int s;

//    *(void **) (&libc_socket) = dlsym(RTLD_NEXT, "socket");
    libc_socket = dlsym(RTLD_NEXT, "socket");
    if (dlerror()) {
        errno = EACCES;
        return -1;
    }

//    s = (*libc_socket)(domain, type, protocol);
    s = libc_socket(domain, type, protocol);
    
    if (s != -1)
        fdOpen[s] = -1;

    if (filefd == 0)
        reportLocation();  // open pipe or log file soon

    /*
        //---------------------------
        char toadd[MAX_BUF_SIZE];
        memset(toadd, 0, sizeof(toadd));
        sprintf(toadd, "sock fd->%d\n", s);
        write(filefd, toadd, sizeof(toadd));
        //---------------------------
     */
    return s;
}


int bind(int sockfd, const struct sockaddr *my_addr, socklen_t addrlen);

int bind(int sockfd, const struct sockaddr *my_addr, socklen_t addrlen) {
    int (*libc_bind)(int, const struct sockaddr*, int);
    int s;

//    *(void **) (&libc_bind) = dlsym(RTLD_NEXT, "bind");
    libc_bind = dlsym(RTLD_NEXT, "bind");
    if (dlerror()) {
        errno = EACCES;
        return -1;
    }

//    s = (*libc_bind)(sockfd, my_addr, addrlen);
    s = libc_bind(sockfd, my_addr, addrlen);
    
    if (s != -1 && fdOpen[sockfd] <= 0) {
        //check what was the port associated with socket sockfd:
        struct sockaddr_in after;
        socklen_t len = sizeof ( after);
        getsockname(sockfd, (struct sockaddr *)&after, &len);
        int port = ntohs(((struct sockaddr_in *) &after)->sin_port);

        //writes the port
        if (port > 0) {
            writeSocketPort(1, port);
            fdOpen[sockfd] = port;
        }
    }
    /*
        //---------------------------
        char toadd[MAX_BUF_SIZE];
        memset(toadd, 0, sizeof(toadd));
        sprintf(toadd, "bind fd->%d\n", sockfd);
        write(filefd, toadd, sizeof(toadd));
        //---------------------------
     */
    return s;
}

int connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen);

int connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen) {
    int (*libc_connect)(int, const struct sockaddr*, socklen_t);
    int s;

//    *(void **) (&libc_connect) = dlsym(RTLD_NEXT, "connect");
    libc_connect = dlsym(RTLD_NEXT, "connect");
    if (dlerror()) {
        errno = EACCES;
        return -1;
    }

//    s = (*libc_connect)(sockfd, serv_addr, addrlen);
    s = libc_connect(sockfd, serv_addr, addrlen);

    if (s != -1 && fdOpen[sockfd] <= 0) {
        //check what was the port associated with socket sockfd:
        struct sockaddr_in after;
        socklen_t len = sizeof ( after);
        getsockname(sockfd, (struct sockaddr *)&after, &len);
        int port = ntohs(((struct sockaddr_in *) &after)->sin_port);

        //writes the port
        if (port > 0) {
            writeSocketPort(1, port);
            fdOpen[sockfd] = port;
        }
    }


    /*
        //---------------------------
        char toadd[MAX_BUF_SIZE];
        memset(toadd, 0, sizeof(toadd));
        sprintf(toadd, "connect fd->%d\n", sockfd);
        write(filefd, toadd, sizeof(toadd));
        //---------------------------
     */
    return s;
}

//-----------------------------------------
int accept(int sockfd, struct sockaddr *addr, socklen_t *addrlen);

int accept(int sockfd, struct sockaddr *addr, socklen_t *addrlen) {
    int (*libc_accept)(int, const struct sockaddr*, socklen_t*);
    int s;

//    *(void **) (&libc_accept) = dlsym(RTLD_NEXT, "accept");
    libc_accept = dlsym(RTLD_NEXT, "accept");
    if (dlerror()) {
        errno = EACCES;
        return -1;
    }

//    s = (*libc_accept)(sockfd, addr, addrlen);
    s = libc_accept(sockfd, addr, addrlen);
    
    if (fdOpen[sockfd] <= 0) {
        //check what was the port associated with socket sockfd:
        struct sockaddr_in after;
        socklen_t len = sizeof ( after);
        getsockname(s, (struct sockaddr *)&after, &len);
        int port = ntohs(((struct sockaddr_in *) &after)->sin_port);

        //writes the port
        if (port > 0) {
            writeSocketPort(1, port);
            fdOpen[sockfd] = port;
        }
    }
    if (s != -1) {
        int sockRep = fdOpen[sockfd];
        writeSocketPort(1, sockRep);
        fdOpen[s] = sockRep;
    }


    /*
        //---------------------------
        char toadd[MAX_BUF_SIZE];
        memset(toadd, 0, sizeof(toadd));
        sprintf(toadd, "accept in %d fd->%d\n", sockfd, s);
        write(filefd, toadd, sizeof(toadd));
        //---------------------------
     */
    return s;
}


int listen(int sockfd, int backlog);

int listen(int sockfd, int backlog) {
    int (*libc_listen)(int, int);
    int s;

//    *(void **) (&libc_listen) = dlsym(RTLD_NEXT, "listen");
    libc_listen = dlsym(RTLD_NEXT, "listen");
    if (dlerror()) {
        errno = EACCES;
        return -1;
    }

//    s = (*libc_listen)(sockfd, backlog);
    s = libc_listen(sockfd, backlog);
    /*
        //---------------------------
        char toadd[MAX_BUF_SIZE];
        memset(toadd, 0, sizeof(toadd));
        sprintf(toadd, "listen fd->%d\n", sockfd);
        write(filefd, toadd, sizeof(toadd));
        //---------------------------
     */
    return s;
}

//-----------------------------------------

int close(int fd);

int close(int fd) {

    int (*libc_close)(int);
    int s;

//    *(void **) (&libc_close) = dlsym(RTLD_NEXT, "close");
    libc_close = dlsym(RTLD_NEXT, "close");
    if (dlerror()) {
        errno = EACCES;
        return -1;
    }

//    s = (*libc_close)(fd);
    s = libc_close(fd);
    if (s != -1 && fdOpen[fd] > 0) {
        int port = fdOpen[fd];
        writeSocketPort(0, port);
        fdOpen[fd] = 0;
    }
    /*
        //---------------------------
        char toadd[MAX_BUF_SIZE];
        memset(toadd, 0, sizeof(toadd));
        sprintf(toadd, "close fd->%d\n", fd);
        write(filefd, toadd, sizeof(toadd));
        //---------------------------
     */
    return s;
}


ssize_t send(int s, const void *buf, size_t len, int flags);

ssize_t send(int s, const void *buf, size_t len, int flags) {
    int (*libc_send)(int, const void *, size_t, int);
    int r;

//    *(void **) (&libc_send) = dlsym(RTLD_NEXT, "send");
    libc_send = dlsym(RTLD_NEXT, "send");
    if (dlerror()) {
        errno = EACCES;
        return -1;
    }

//    r = (*libc_send)(s, buf, len, flags);
    r = libc_send(s, buf, len, flags);

    if (r != -1 && fdOpen[s] <= 0) {
        //check what was the port associated with socket sockfd:
        struct sockaddr_in after;
        socklen_t len = sizeof ( after);
        getsockname(s, (struct sockaddr *)&after, &len);
        int port = ntohs(((struct sockaddr_in *) &after)->sin_port);

        //writes the port
        if (port > 0) {
            writeSocketPort(1, port);
            fdOpen[s] = port;
        }
    }
    /*
        //---------------------------
        char toadd[MAX_BUF_SIZE];
        memset(toadd, 0, sizeof(toadd));
        sprintf(toadd, "sendto fd->%d", s);
        write(filefd, toadd, sizeof(toadd));
        //---------------------------
     */
    return r;
}

ssize_t write(int fd, const void *buf, size_t count);

ssize_t write(int fd, const void *buf, size_t count) {
    int (*libc_write)(int, const void *, size_t);
    int r;

//    *(void **) (&libc_write) = dlsym(RTLD_NEXT, "write");
    libc_write = dlsym(RTLD_NEXT, "write");
    if (dlerror()) {
        errno = EACCES;
        return -1;
    }

//    r = (*libc_write)(fd, buf, count);
    r = libc_write(fd, buf, count);

    if (r != -1 && fdOpen[fd] <= 0) {
        //check what was the port associated with socket sockfd:
        struct sockaddr_in after;
        socklen_t len = sizeof ( after);
        getsockname(fd, (struct sockaddr *)&after, &len);
        int port = ntohs(((struct sockaddr_in *) &after)->sin_port);

        //writes the port
        if (port > 0) {
            writeSocketPort(1, port);
            fdOpen[fd] = port;
        }
    }

    /*
        //---------------------------
        char toadd[MAX_BUF_SIZE];
        memset(toadd, 0, sizeof(toadd));
        sprintf(toadd, "write fd->%d\n", fd);
        (*libc_write)(filefd, toadd, sizeof(toadd));
        //---------------------------
     */
    return r;
}

ssize_t sendto(int s, const void *buf, size_t len, int flags, const struct sockaddr *to, socklen_t tolen);

ssize_t sendto(int s, const void *buf, size_t len, int flags, const struct sockaddr *to, socklen_t tolen) {
    int (*libc_sendto)(int, const void *, size_t, int, const struct sockaddr *, socklen_t);
    int r;

//    *(void **) (&libc_sendto) = dlsym(RTLD_NEXT, "sendto");
    libc_sendto = dlsym(RTLD_NEXT, "sendto");
    if (dlerror()) {
        errno = EACCES;
        return -1;
    }

//    r = (*libc_sendto)(s, buf, len, flags, to, tolen);
    r = (*libc_sendto)(s, buf, len, flags, to, tolen);
    if (r != -1 && fdOpen[s] <= 0) {
        //check what was the port associated with socket sockfd:
        struct sockaddr_in after;
        socklen_t len = sizeof ( after);
        getsockname(s, (struct sockaddr *)&after, &len);
        int port = ntohs(((struct sockaddr_in *) &after)->sin_port);

        //writes the port
        if (port > 0) {
            writeSocketPort(1, port);
            fdOpen[s] = port;
        }
    }
    
    /*
        //---------------------------
        char toadd[MAX_BUF_SIZE];
        memset(toadd, 0, sizeof(toadd));
        sprintf(toadd, "sendto fd->%d\n", s);
        write(filefd, toadd, sizeof(toadd));
        //---------------------------
     */
    return r;
}

/*
ssize_t sendmsg(int s, const struct msghdr *msg, int flags);
ssize_t sendmsg(int s, const struct msghdr *msg, int flags)
{
    int (*libc_sendmsg)(int, const struct msghdr *, int);
    int r;

 *(void **)(&libc_sendmsg) = dlsym(RTLD_NEXT, "sendmsg");
    if(dlerror()) {
        errno = EACCES;
        return -1;
    }

    r = (*libc_sendmsg)(s, msg, flags);

    //---------------------------
    char toadd[MAX_BUF_SIZE];
    memset(toadd, 0, sizeof(toadd));
    sprintf(toadd, "sendmsg fd->%d\n", s);
    write(filefd, toadd, sizeof(toadd));
    //---------------------------

    return r;
}
 */
