#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include "procMonitor.h"


//----------GLOBAL VARIABLES----------
extern char* pipeFile;

#define MAX_BUF_SIZE	255
#define MAX_SOCKETS		50

extern int toexit;

int sockets[65535];
int pipefd;
int currcounter;
//------------------------------------

//estrutura para guardar ips/pids
typedef struct{

	int ip[4];
	int port;
	int inode;
	char type[10]; //tcp, tcp6, udp, unix

}IpDataStr, *ipdata;

typedef struct{

	int pid;
	ipdata sockets[MAX_SOCKETS]; //limitado a monitorização de 500 sockets???
	int nrentrys;

}ProcessSocketsStr, *psockets;

psockets createProcessSockets(int pid){

	psockets ps = (psockets)malloc(sizeof(ProcessSocketsStr));
	ps->pid = pid;
	ps->nrentrys = 0;
	return ps;
}

void freePSockets(psockets ps)
{
	int i = 0;
	for(i = 0; i < ps->nrentrys; i++)
	{
		free(ps->sockets[i]);
	}
	free(ps);
}

int inodeSocket(psockets ps, int in){

	int i;
	ipdata id;

	for(i = 0; i < ps->nrentrys; i++)
	{
		id = ps->sockets[i];

		if(id->inode == in)
		{
			return id->port;
		}
	}

	return -1; //devolve erro
}

void hexToDec(const char * d, int length, char* cres){

	int result = 0;
	int i;
	char tmp;

	for(i = 0; i < length; i++)
	{
		if(d[i] == 'A')
			result += 10 * pow(16, length -1 - i);
		else
			if(d[i] == 'B')
				result += 11 * pow(16, length -1 - i);
		else
			if(d[i] == 'C')
				result += 12 * pow(16, length -1 - i);
		else
			if(d[i] == 'D')
				result += 13 * pow(16, length -1 - i);
		else
			if(d[i] == 'E')
				result += 14 * pow(16, length -1 - i);
		else
			if(d[i] == 'F')
				result += 15 * pow(16, length -1 - i);
		else
		{
			tmp = d[i];
			result += atoi(&tmp) * pow(16, length -1 - i);
		}
	}

	sprintf(cres, "%d", result);
}

void ipParsing(const char* raw, ipdata ipd){

	char ippart1[5];
	char ippart2[5];
	char ippart3[5];
	char ippart4[5];
	char port[7];
if(strlen(raw) < 30)
{
	ippart1[0] = raw[0];
	ippart1[1] = raw[1];
	ippart1[2] = '\0';
	ippart2[0] = raw[2];
	ippart2[1] = raw[3];
	ippart2[2] = '\0';
	ippart3[0] = raw[4];
	ippart3[1] = raw[5];
	ippart3[2] = '\0';
	ippart4[0] = raw[6];
	ippart4[1] = raw[7];
	ippart4[2] = '\0';
	port[0] = raw[9];
	port[1] = raw[10];
	port[2] = raw[11];
	port[3] = raw[12];
	port[4] = '\0';

	hexToDec(ippart1, 2, ippart1);
	hexToDec(ippart2, 2, ippart2);
	hexToDec(ippart3, 2, ippart3);
	hexToDec(ippart4, 2, ippart4);
	hexToDec(port, 4, port);

	ipd->ip[0] = atoi(ippart4);
	ipd->ip[1] = atoi(ippart3);
	ipd->ip[2] = atoi(ippart2);
	ipd->ip[3] = atoi(ippart1);
	ipd->port = atoi(port);
}
else
{
	ippart1[0] = raw[24];
	ippart1[1] = raw[25];
	ippart1[2] = '\0';
	ippart2[0] = raw[26];
	ippart2[1] = raw[27];
	ippart2[2] = '\0';
	ippart3[0] = raw[28];
	ippart3[1] = raw[29];
	ippart3[2] = '\0';
	ippart4[0] = raw[30];
	ippart4[1] = raw[31];
	ippart4[2] = '\0';
	port[0] = raw[33];
	port[1] = raw[34];
	port[2] = raw[35];
	port[3] = raw[36];
	port[4] = '\0';

	hexToDec(ippart1, 2, ippart1);
	hexToDec(ippart2, 2, ippart2);
	hexToDec(ippart3, 2, ippart3);
	hexToDec(ippart4, 2, ippart4);
	hexToDec(port, 4, port);

	ipd->ip[0] = atoi(ippart4);
	ipd->ip[1] = atoi(ippart3);
	ipd->ip[2] = atoi(ippart2);
	ipd->ip[3] = atoi(ippart1);
	ipd->port = atoi(port);
}
}

void getNetInfo(psockets ps){

	//tcp getinfo
	FILE *file = fopen("/proc/net/tcp", "r");

	char buf[150];
	int i = 0; //for socket control

	//data readed fr	om tcp
	int nrline;
	char hostip[50];
	char remip[50];
	char st[5];
	char tx[20];
	char tr[20];
	char ret[20];
	int uid;
	int timeout;
	int inode;

	if(fgets(buf, 150, file) == NULL)
		perror("PROC MONITOR GETNETINFO");

	while( (fgets(buf, 150, file)) != NULL && ps->nrentrys + 1 < MAX_SOCKETS)
	{

		if(strlen(buf) > 20)
		{
     		sscanf(buf, "%d: %s %s %s %s %s %s %d %d %d",&nrline, hostip, remip, st, tx, tr, ret, &uid, &timeout, &inode);

			ps->sockets[i] = (ipdata)malloc(sizeof(IpDataStr));
			ipParsing(hostip, ps->sockets[i]);
			ps->sockets[i]->inode = inode;

			ps->nrentrys = ps->nrentrys + 1;
			strcpy(ps->sockets[i]->type, "tcp");
			i++;
		}
	}

	fclose(file);

	//tcp6 getinfo
	file = fopen("/proc/net/tcp6", "r");

	if(fgets(buf, 150, file) == NULL)
		perror("PROC MONITOR GETNETINFO");

	while( (fgets(buf, 150, file)) != NULL && ps->nrentrys + 1 < MAX_SOCKETS)
	{
		if(strlen(buf) > 20)
		{
     		sscanf(buf, "%d: %s %s %s %s %s %s %d %d %d",&nrline, hostip, remip, st, tx, tr, ret, &uid, &timeout, &inode);

			ps->sockets[i] = (ipdata)malloc(sizeof(IpDataStr));
			ipParsing(hostip, ps->sockets[i]);
			ps->sockets[i]->inode = inode;
			ps->nrentrys = ps->nrentrys + 1;
			strcpy(ps->sockets[i]->type, "tcp");
			i++;
		}
	}
	fclose(file);

	//udp getinfo
	file = fopen("/proc/net/udp", "r");

	if(fgets(buf, 150, file) == NULL)
		perror("PROC MONITOR GETNETINFO 2");

	while( (fgets(buf, 150, file)) != NULL  && ps->nrentrys + 1< MAX_SOCKETS)
	{
		if(strlen(buf) > 20)
		{
     		sscanf(buf, "%d: %s %s %s %s %s %s %d %d %d",&nrline, hostip, remip, st, tx, tr, ret, &uid, &timeout, &inode);
			//printf("ip:%s -> dest:%s -> uid:%d -> inode:%d\n", hostip, remip, uid, inode);

			ps->sockets[i] = (ipdata)malloc(sizeof(IpDataStr));
			ipParsing(hostip, ps->sockets[i]);
			ps->sockets[i]->inode = inode;
			ps->nrentrys = ps->nrentrys + 1;
			strcpy(ps->sockets[i]->type, "udp");
			i++;
		}
	}
	
	fclose(file);

//udp6 getinfo
	file = fopen("/proc/net/udp6", "r");

	if(fgets(buf, 150, file) == NULL)
		perror("PROC MONITOR GETNETINFO 2");

	while( (fgets(buf, 150, file)) != NULL  && ps->nrentrys + 1< MAX_SOCKETS)
	{
		if(strlen(buf) > 20)
		{
     		sscanf(buf, "%d: %s %s %s %s %s %s %d %d %d",&nrline, hostip, remip, st, tx, tr, ret, &uid, &timeout, &inode);
			//printf("ip:%s -> dest:%s -> uid:%d -> inode:%d\n", hostip, remip, uid, inode);

			ps->sockets[i] = (ipdata)malloc(sizeof(IpDataStr));
			ipParsing(hostip, ps->sockets[i]);
			ps->sockets[i]->inode = inode;
			ps->nrentrys = ps->nrentrys + 1;
			strcpy(ps->sockets[i]->type, "udp");
			i++;
		}
	}
	fclose(file);

	
}

void listDirectorySockets(int pid){

	DIR *d;
	struct dirent *dir;
	struct stat filep;
	char buf[65535];
	char fullpath[100];
	char* name;
	char toadd[50];
	int sockinode;

	char path[100];
	//char filter[200];
	//memset(filter, 0, sizeof(filter));

	int res;

	sprintf(path, "/proc/%d/fd", pid);

	psockets ps = createProcessSockets(pid); //mudar o pid do processo
	getNetInfo(ps); //obtem informação sobre sockets /proc/net

	d = opendir(path);
	if(d != NULL)
	{
		while( (dir = readdir(d)) != NULL)
		{
			name = dir->d_name;
			if(strcmp(name, ".") != 0 && strcmp(name, "..") != 0)
			{
				strcpy(fullpath, path);
				strcat(fullpath, "/");
				strcat(fullpath, name);

				stat(fullpath, &filep);

				if(S_ISSOCK(filep.st_mode) != 0)
				{
					readlink(fullpath, buf, 65535);
					sscanf(buf, "socket:[%d]", &sockinode);

					res = inodeSocket(ps, sockinode);
					if(res > 0 && res < 65535)
					{
						if(sockets[res] == 0)
						{
							memset(toadd, 0, sizeof(toadd));
							sprintf(toadd, "%d %d\n", 1, res);
							write(pipefd, toadd, sizeof(toadd));
							sockets[res] = currcounter;
						}
						else
						{
							sockets[res] = currcounter;
						}
					}
				}
			}
		}
	}

	freePSockets(ps);
}

void checkUnusedSockets(){
	int i;
	for(i = 0; i < 65535; i++)
	{
		if(sockets[i] != 0 && sockets[i] != currcounter)
		{
			//terminado
			char toadd[50];
			memset(toadd, 0, sizeof(toadd));
			sprintf(toadd, "%d %d\n", 0, i);
			write(pipefd, toadd, sizeof(toadd));
			sockets[i] = 0;
		}
	}
}

int findPidBySocket(int sock){

	int i;
	int inode = 0;
	DIR* d;
	DIR* e;
	int sockinode;
	char buf[65535];
	char path[100];
	char fullpath[100];
	char* namepid;
	char* name;
	struct stat filep;
	struct dirent *piddir;
	struct dirent *fddir;
	psockets ps = createProcessSockets(0); //pid = 0 pois não se sabe qual é

	getNetInfo(ps); //obtem informação sobre sockets /proc/net

	for(i = 0; i < ps->nrentrys; i++)
	{
		//verifica se encontra o fd ligado ao socket com o porto sock
		if(ps->sockets[i]->port == sock)
		{
			inode = ps->sockets[i]->inode;
			break;
		}
	}

	if(inode != 0)
	{
		//percorre todas as directorias /proc
		d = opendir("/proc");
		if(d != NULL)
		{
			while( (piddir = readdir(d)) != NULL)
			{
				namepid = piddir->d_name;
				if(strcmp(namepid, ".") != 0 && strcmp(namepid, "..") != 0)
				{
					memset(path, 0, sizeof(path));
					sprintf(path, "/proc/%s/fd", namepid);

					//le cada entrada da /proc/<pid>/fd
					e = opendir(path);
					if(e != NULL)
					{
						while( (fddir = readdir(e)) != NULL)
						{
							name = fddir->d_name;
							if(strcmp(name, ".") != 0 && strcmp(name, "..") != 0)
							{
								strcpy(fullpath, path);
								strcat(fullpath, "/");
								strcat(fullpath, name);

								stat(fullpath, &filep);

								if(S_ISSOCK(filep.st_mode) != 0)
								{
									readlink(fullpath, buf, 65535);
									sscanf(buf, "socket:[%d]", &sockinode);
									if(sockinode == inode)
									{
										freePSockets(ps);
										return atoi(namepid);
									}
								}
							}
						}
					}
				}
			}
		}
	}
	freePSockets(ps);
	return 0;
}

int msleep(unsigned long milisec)
{
	struct timespec req={0};
	time_t sec=(int)(milisec/1000);
	milisec=milisec-(sec*1000);
	req.tv_sec=sec;
	req.tv_nsec=milisec*1000000L;
	while(nanosleep(&req,&req)==-1)
		continue;
	return 1;
}

void* pidMonitorThread(void * pid){

	pipefd = open(pipeFile, O_WRONLY);
	currcounter = 1;

	while(1){
		msleep(1000);
		if(toexit == 1)
			break;

		listDirectorySockets((int)pid); //gives warning in 64bits
		checkUnusedSockets();
		currcounter++;
		if(currcounter == 1000)
			currcounter = 1;
	}
	printf("EXIT THREAD procMonitor\n");
	return 0;
}

void* socketMonitorThread(void * socket){

	int pid = findPidBySocket((int)socket); //gives warning in 64bits
	if(pid != 0)
		pidMonitorThread((void*)pid); //gives warning in 64bits
	return 0;
}


